#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "tcl.h"
#include "tk.h"
#include "rtp.h"
#include "version.h"

#ifdef WIN32
#define WS_VERSION_ONE MAKEWORD(1,1)
#define WS_VERSION_TWO MAKEWORD(2,2)
#endif

extern char 	 ui[];
Tcl_Interp	*interp;

static int tcl_init(int argc, char **argv)
{
	char		*cmd_line_args, buffer[10];
	Tcl_Obj 	*ui_obj;

	Tcl_FindExecutable(argv[0]);
	interp        = Tcl_CreateInterp();
	cmd_line_args = Tcl_Merge(argc - 1, argv + 1);
	Tcl_SetVar(interp, "argv", cmd_line_args, TCL_GLOBAL_ONLY);

	sprintf(buffer, "%d", argc - 1);
	Tcl_SetVar(interp, "argc", buffer, TCL_GLOBAL_ONLY);
	Tcl_SetVar(interp, "argv0", argv[0], TCL_GLOBAL_ONLY);
	Tcl_SetVar(interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);

	if (Tcl_Init(interp) != TCL_OK) {
                debug_msg("%s\n", Tcl_GetStringResult(interp));
                return FALSE;
        }
        if (Tk_Init(interp) != TCL_OK) {
                debug_msg("%s\n", Tcl_GetStringResult(interp));
                return FALSE;
        }

	ui_obj = Tcl_NewStringObj(ui, strlen(ui));
	if (Tcl_EvalObj(interp, ui_obj) != TCL_OK) {
		debug_msg("ui error: %s\n", Tcl_GetStringResult(interp));
                return FALSE;
	}

	while (Tcl_DoOneEvent(TCL_DONT_WAIT | TCL_ALL_EVENTS)) {
		/* Processing Tcl events, to allow the UI to initialize... */
	};

	Tcl_ResetResult(interp);
	return TRUE;
}

static void tcl_send(char *command)
{
	/* This is called to send a message to the user interface...  */
	/* If the UI is not enabled, the message is silently ignored. */
	assert(command != NULL);

	if (Tk_GetNumMainWindows() <= 0) {
		return;
	}

	if (Tcl_Eval(interp, command) != TCL_OK) {
		debug_msg("TCL error: %s\n", Tcl_GetVar(interp, "errorInfo", 0));
	}
}

static void rtp_callback(struct rtp *s, rtp_event *e)
{
	rtp_packet	*p    = (rtp_packet *)     e->data;
	rtcp_sr         *sr   = (rtcp_sr *)        e->data;
	rtcp_rr         *rr   = (rtcp_rr *)        e->data;
	rtcp_sdes_item	*sdes = (rtcp_sdes_item *) e->data;
	char		 cmd[1024];

	UNUSED(sr);

	switch (e->type) {
		case RX_RTP:
			sprintf(cmd, "set CODEC(0x%08lx) %d", p->ssrc, p->pt);
			tcl_send(cmd);
			xfree(p);
			break;
		case RX_RTCP_START:
			break;
		case RX_RTCP_FINISH:
			break;
		case RX_SR:
			break;
		case RX_RR:
			sprintf(cmd, "chart_set 0x%08lx 0x%08lx %d", rr->ssrc, e->ssrc, (100 * rr->fract_lost) / 256);
			tcl_send(cmd);
			break;
		case RX_RR_EMPTY:
			sprintf(cmd, "rr_empty 0x%08lx", e->ssrc);
			tcl_send(cmd);
			break;
		case RX_SDES:
			switch (sdes->type) {
				case RTCP_SDES_CNAME:
					sprintf(cmd, "set_cname 0x%08lx \"%s\"", e->ssrc, rtp_get_sdes(s, e->ssrc, RTCP_SDES_CNAME));
					tcl_send(cmd);
					break;
				case RTCP_SDES_NAME:
					sprintf(cmd, "set_name 0x%08lx \"%s\"", e->ssrc, rtp_get_sdes(s, e->ssrc, RTCP_SDES_NAME));
					tcl_send(cmd);
					break;
				case RTCP_SDES_EMAIL:
					sprintf(cmd, "set_email 0x%08lx \"%s\"", e->ssrc, rtp_get_sdes(s, e->ssrc, RTCP_SDES_EMAIL));
					tcl_send(cmd);
					break;
				case RTCP_SDES_PHONE:
					sprintf(cmd, "set_phone 0x%08lx \"%s\"", e->ssrc, rtp_get_sdes(s, e->ssrc, RTCP_SDES_PHONE));
					tcl_send(cmd);
					break;
				case RTCP_SDES_LOC:
					sprintf(cmd, "set_loc 0x%08lx \"%s\"", e->ssrc, rtp_get_sdes(s, e->ssrc, RTCP_SDES_LOC));
					tcl_send(cmd);
					break;
				case RTCP_SDES_TOOL:
					sprintf(cmd, "set_tool 0x%08lx \"%s\"", e->ssrc, rtp_get_sdes(s, e->ssrc, RTCP_SDES_TOOL));
					tcl_send(cmd);
					break;
				case RTCP_SDES_NOTE:
					break;
				default:
					break;
			}
			break;
		case RX_BYE:
			sprintf(cmd, "ssrc_delete 0x%08lx", e->ssrc);
			tcl_send(cmd);
			break;
		case SOURCE_CREATED:
			sprintf(cmd, "ssrc_exists 0x%08lx", e->ssrc);
			tcl_send(cmd);
			break;
		case SOURCE_DELETED:
			sprintf(cmd, "ssrc_delete 0x%08lx", e->ssrc);
			tcl_send(cmd);
			break;
		case RR_TIMEOUT:
			sprintf(cmd, "chart_set 0x%08lx 0x%08lx %d", rr->ssrc, e->ssrc, 101);
			tcl_send(cmd);
			break;
		default:
			debug_msg("Unknown event: type %d ssrc 0x%08lx\n", e->type, e->ssrc);
			break;
	}
}

static double tv_diff(struct timeval curr_time, struct timeval prev_time)
{
    /* Return curr_time - prev_time */
    double sec;
    double usec;

    sec = curr_time.tv_sec - prev_time.tv_sec ;
    usec = curr_time.tv_usec - prev_time.tv_usec ;

    assert(sec >= 0);
    if (usec < 0 ) {
	assert(sec > 0);
	sec--;
	usec += 1000000;
    }

    return (sec+(usec/1000000));
}

int main(int argc, char *argv[])
{
	struct timeval	 timeout, curr_t, prev_t;
	struct rtp	*rtp_session;
	char		*addr;
	char		*port;
	char		 cmd[100];
	u_int32		 curr_rtp_time;
#ifdef WIN32
	WSADATA WSAdata;
	if (WSAStartup(WS_VERSION_TWO, &WSAdata) != 0 && WSAStartup(WS_VERSION_ONE, &WSAdata) != 0) {
    		printf("Windows Socket initialization failed.\n");
		return 1;
	}
	debug_msg("WSAStartup OK: %sz\nStatus:%s\n", WSAdata.szDescription, WSAdata.szSystemStatus);
#endif

	if (argc != 2) {
		printf("Usage examples: rqm 224.2.186.225/26406\n");
		printf("                rqm -version\n");
		return 1;
	}
	if (strcmp(argv[1], "-version") == 0) {
		printf("%s\n", VERSION);
		return 1;
	}
	addr = (char *) strtok(argv[1], "/");
	port = (char *) strtok(NULL, "\0");

	rtp_session = rtp_init(addr, atoi(port), atoi(port), 15, 64000, rtp_callback, NULL);

	/* Initialise the user interface... */
	if (tcl_init(argc, argv) == FALSE) {
		printf("Unable to initialise interface\n");
		return 1;
	}
	sprintf(cmd, "ssrc_exists 0x%08lx", rtp_my_ssrc(rtp_session));
	tcl_send(cmd);
	sprintf(cmd, "set group_addr %s\n", addr);
	tcl_send(cmd);
	sprintf(cmd, "wm title . \"Reception quality matrix %s/%d\"", addr, atoi(port));
	tcl_send(cmd);
	rtp_set_sdes(rtp_session, rtp_my_ssrc(rtp_session), RTCP_SDES_TOOL, VERSION, strlen(VERSION));

	gettimeofday(&prev_t, NULL);
	curr_rtp_time = (u_int32) lrand48();
	while (Tk_GetNumMainWindows() > 0) {
		gettimeofday(&curr_t, NULL);
		curr_rtp_time += (int) (tv_diff(curr_t, prev_t) * 8000);
		prev_t = curr_t;
		timeout.tv_sec  = 0;
		timeout.tv_usec = 100000;
		rtp_recv(rtp_session, &timeout, curr_rtp_time);
		rtp_send_ctrl(rtp_session, curr_rtp_time, NULL);
		rtp_update(rtp_session);
		while (Tcl_DoOneEvent(TCL_DONT_WAIT | TCL_ALL_EVENTS)) {
			/* Process Tcl events... */
		};
	}
	rtp_send_bye(rtp_session);
	rtp_done(rtp_session);
#ifdef WIN32
	WSACleanup();
#endif
	return 0;
}

